/*
 ******************************************************************
 * HISTORY
 * 18-Feb-07  J.Cooperstock, McGill University, for stock market assignment
 * 07-Mar-02  J. Cooperstock, McGill University, for face recognition assignment
 * 15-Oct-94  Jeff Shufelt (js), Carnegie Mellon University
 *
 ******************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "backprop.h"
#include "config.h"

/* math macros */
#define SQ(x) ((x) * (x))

/* storage of data vectors */
typedef struct { 
  int n;  
  double input[MAX_DATA][N_INPUT];
  double target[MAX_DATA][N_OUTPUT];
} DATASET;

double performance_on_data();
void backprop (DATASET *, DATASET *, int , int , char *, double, double );
void user_test(BPNN *);



main(int argc, char *argv[]) {
  char *netname=NULL, *trainname=NULL, *testname=NULL;
  DATASET *trainset, *testset;
  int ind, epochs = 100, seed = rand(), savedelta = 100;
  double alpha = DEF_ALPHA, gamma = DEF_GAMMA;

  if (argc < 2) {
    usage(argv[0]);
    exit (-1);
  }

  /*** Create vectors ***/
  trainset = (DATASET *) malloc (sizeof (DATASET));
  testset = (DATASET *) malloc (sizeof (DATASET));

  /*** Scan command line ***/
  for (ind = 1; ind < argc; ind++) {

    /*** Parse switches ***/
    if (argv[ind][0] == '-') {
      switch (argv[ind][1]) {  
	case 'a': 
	  alpha = atof(argv[++ind]);
	  fprintf (stderr, "Using alpha (learning rate) of %lf\n", alpha);
	  break;
	case 'g': 
	  gamma = atof(argv[++ind]);
	  fprintf (stderr, "Using gamma (learning rate) of %lf\n", gamma);
	  break;
        case 'n': 
	  netname = argv[++ind];
	  break;
        case 'e': 
	  epochs = atoi(argv[++ind]);
	  break;
        case 's': 
	  seed = atoi(argv[++ind]);
	  break;
        case 'S': 
	  savedelta = atoi(argv[++ind]);
	  break;
        case 't': 
	  trainname = argv[++ind];
	  break;
        case 'v': 
	  testname =  argv[++ind];
	  break;
        default : 
	  usage(argv[0]);
	  exit(1);
      }
    }
  }

  /*** Sanity check ***/
  if (!trainname || !testname || !netname) {
    usage(argv[0]);
    exit(1);
  }

  load_vectors_from_textfile (trainset, trainname);
  load_vectors_from_textfile (testset, testname);

  /*** Display counts **/
  fprintf (stderr, "%d vectors in train set\n", trainset->n);
  fprintf (stderr, "%d vectors in test set\n", testset->n);

  /*** Initialize the neural net package ***/
  bpnn_initialize(seed);

  /*** train the network using backpropagation ***/
  backprop(trainset, testset, epochs, savedelta, netname, alpha, gamma);
}


load_vectors_from_textfile (DATASET *dataset, char *filename) {
  FILE *fp;
  int i, j, s, n;
  double *in, *out;

  fp = fopen (filename, "r");
  fscanf (fp, "%d", &n);
  dataset->n = n;

  for (s = 0; s < n; s++) {
  	in = dataset->input[s];
  	out = dataset->target[s];
        for (i = 0; i < N_INPUT; i++)
		scanf ("%lf", in+i); 
        for (j = 0; j < N_OUTPUT; j++)
		scanf ("%lf", out+j); 
  }

  fclose (fp);
}

void backprop (DATASET *traindata, DATASET *testdata, int epochs, int savedelta, char *netname, double alpha, double gamma) {
  BPNN *net;
  int epoch, i, j;
  double out_err, hid_err, sumerr;
  int train_n = traindata->n;

  /*** Read network in if it exists, otherwise make one from scratch ***/
  if ((net = bpnn_read(netname)) == NULL) {
    if (train_n > 0) {
      printf("Creating new network '%s'\n", netname);
      net = bpnn_create(N_INPUT, N_HIDDEN, N_OUTPUT);
    } else {
      printf("Need some images to train on, use -t\n");
      return;
    }
  }

  if (epochs > 0) {
    printf("Training underway (going to %d epochs)\n", epochs);
    printf("Will save network every %d epochs\n", savedelta);
    fflush(stdout);
  }

  /*** Print out performance before any epochs have been completed. ***/
  printf("0\t");  fflush(stdout);
  printf ("training data=%lf\t", performance_on_data(net, traindata));
  printf ("test data=%lf\n", performance_on_data(net, testdata));

  /************** Train it *****************************/
  for (epoch = 1; epoch <= epochs; epoch++) {

    printf("%d\t", epoch);  fflush(stdout);

    sumerr = 0.0;
    for (i = 0; i < train_n; i++) {

      /** Provide training input to network **/
	for (j = 0; j < N_INPUT; j++)
      		net->input_units[j+1] = traindata->input[i][j];

      /** Specify training value target outputs for network **/
	for (j = 0; j < N_OUTPUT; j++)
      		net->target[j+1] = traindata->target[i][j];

      /** Run backprop with learning rate alpha and momentum gamma **/
      bpnn_train(net, alpha, gamma, &out_err, &hid_err);

      sumerr += (out_err + hid_err);
    }
    /* printf("t_err:%g\t", sumerr); */

    /*** Evaluate performance on train, test, test2, and print perf ***/
    printf ("training data=%lf\t", performance_on_data(net, traindata));
    printf ("test data=%lf\n", performance_on_data(net, testdata));

    /*** Save network every 'savedelta' epochs ***/
    if (!(epoch % savedelta)) {
      bpnn_save(net, netname);
    }

  }
  printf("\n"); fflush(stdout);

  /** Save the trained network **/
  if (epochs > 0)
    bpnn_save(net, netname);

  /*** The user would loikely, at this point, want to query the network ***/
  user_test (net);
}


/*** Computes the performance of a net on the images in the imagelist. ***/
/*** Prints out the percentage correct on the image set, and the
     average error between the target and the output units for the set. ***/
double performance_on_data (BPNN *net, DATASET *data) {
  double sse, delta;
  int i, j, n;

  sse = 0.0;
  n = data->n;

  if (n > 0) {
    for (i = 0; i < n; i++) {

      /** Provide training input to network **/
      for (j = 0; j < N_INPUT; j++)
         net->input_units[j+1] = data->input[i][j];

      /*** Run the net on this input. **/
      bpnn_feedforward(net);

      /*** calculate sum of squared errors over output units  ***/
      for (j = 0; j < N_OUTPUT; j++)
         sse += SQ(data->target[i][j] - net->output_units[j+1]);
    }

    /*** normalize ***/
    sse = sse/(2*n);
  }

  return (sse);
}

void user_test(BPNN *net) {
  int i, j;

  while (1) {
        printf ("Enter input vector: ");
        for (i = 0; i < N_INPUT; i++)
		scanf ("%lf", &(net->input_units[i+1]));
        bpnn_feedforward(net);
        
        printf ("Calculated output vector: ");
        for (j = 0; j < N_OUTPUT; j++)
		printf ("%lf", net->output_units[j+1]);
	printf ("\n");
   }
}

usage(char *prog) {
  printf("USAGE: %s\n", prog);
  printf("\t[-a <alpha>] -- default %lf\n", DEF_ALPHA);
  printf("\t[-g <gamma>] -- default %lf\n", DEF_GAMMA);
  printf("\t[-e <number of epochs>]\n");
  printf("\t[-s <random number generator seed>]\n");
  printf("\t[-S <number of epochs between saves of network>]\n");
  printf("\t-n <network file>\n");
  printf("\t-t <training set list>\n");
  printf("\t-v <testing set list>\n\n");
}
